/*

Copyright 2019, dettus@dettus.net

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "version.h"
#include "configuration.h"
#include "vm68k.h"
#include "linea.h"
#include "default_callbacks.h"
#define	MAXMAGSIZE	(1<<20)
#define	MAXGFXSIZE	(1<<22)


int dMagnetic_init(void** hVM68k,void** hLineA,void* pSharedMem,int memsize,char* magfilename,char* gfxfilename)
{
	int sizevm68k;
	int sizelineA;
	int retval;
	int version;
	char* magbuf;
	char* gfxbuf;
	int magsize;
	int gfxsize;
	FILE *f;

	// step 1: load the game binaries
	magbuf=malloc(MAXMAGSIZE);
	gfxbuf=malloc(MAXGFXSIZE);
	if (magbuf==NULL || gfxbuf==NULL) 
	{
		fprintf(stderr,"ERROR: unable to allocate memory for the data files\n");
		return -1;
	}
	f=fopen(magfilename,"rb");
	if (f==NULL)
	{
		fprintf(stderr,"ERROR: unable to open [%s]\n",magfilename);
		return -2;
	}
	magsize=fread(magbuf,sizeof(char),MAXMAGSIZE,f);
	fclose(f);
	f=fopen(gfxfilename,"rb");
	if (f==NULL)
	{
		fprintf(stderr,"ERROR: unable to open [%s]\n",gfxfilename);
		return -2;
	}
	gfxsize=fread(gfxbuf,sizeof(char),MAXGFXSIZE,f);
	fclose(f);

	// at this point, they are stored in magbuf and gfxbuf.


	// step 2: start the engine.
	retval=vm68k_getsize(&sizevm68k);
	if (retval) 
	{
		fprintf(stderr,"ERROR: vm68k_getsize returned %d\n",retval);
		return retval;
	}
	retval=lineA_getsize(&sizelineA);
	if (retval) 
	{
		fprintf(stderr,"ERROR: lineA_getsize returned %d\n",retval);
		return retval;
	}
	*hVM68k=malloc(sizevm68k);
	*hLineA=malloc(sizelineA);
	if (*hVM68k==NULL || *hLineA==NULL) 
	{
		fprintf(stderr,"ERROR: unable to allocate memory for the engine\n");
		return -1;
	}
	
	retval=lineA_init(*hLineA,pSharedMem,&memsize,magbuf,magsize,gfxbuf,gfxsize);
	if (retval)
	{
		fprintf(stderr,"ERROR: lineA_init returned %d\n",retval);
		return retval;
	}
	retval=lineA_getVersion(*hLineA,&version);
	if (retval)
	{
		fprintf(stderr,"ERROR: lineA_getversion returned %d\n",retval);
		return retval;
	}
	fprintf(stderr,"Initializing the 68K version %d VM\n",version);	
	retval=vm68k_init(*hVM68k,pSharedMem,memsize,version);
	if (retval)
	{
		fprintf(stderr,"ERROR: vm68k_init returned %d\n",retval);
		return retval;
	}
	
	// the relevant data from the game binaries has been copied.
	free(gfxbuf);
	free(magbuf);
	return 0;
}
int main(int argc,char** argv)
{
	char inifilename[1024];
	char magfilename[1024];
	char gfxfilename[1024];
	int retval;
	FILE *f_inifile=NULL;

	void* hVM68k;
	void* hLineA;
	void* hGUI;
	void* pSharedMem;
	int sharedmemsize;
	int sizeGUI;
	int unknownopcode;


	inifilename[0]=0;
	magfilename[0]=0;
	gfxfilename[0]=0;

	fprintf(stderr,"*** dMagnetic\n");
	fprintf(stderr,"*** Use at your own risk\n");
	fprintf(stderr,"*** (C)opyright 2019 by dettus@dettus.net\n");
	fprintf(stderr,"*****************************************\n");	
	fprintf(stderr,"\n");

	
	if (argc<2)
	{
		fprintf(stderr,"Please run with %s [-ini dMagnetic.ini] GAME\n",argv[0]);
		fprintf(stderr," where GAME is one of\n");
		fprintf(stderr,"    pawn       To play \"The Pawn\"\n");
		fprintf(stderr,"    guild      To play \"The Guild of Thieves\"\n");
		fprintf(stderr,"    jinxter    To play \"Jinxter\"\n");
		fprintf(stderr,"    corruption To play \"Corruption\"\n");
		fprintf(stderr,"    fish       To play \"Fish!\"\n");
		fprintf(stderr,"    myth       To play \"Myth\"\n");
		fprintf(stderr,"    wonderland To play \"Wonderland\"\n");
		return 0;		
	}
	if (retrievefromcommandline(argc,argv,"--version",NULL,0))
	{
		printf("%d.%d%d\n",VERSION_MAJOR,VERSION_MINOR,VERSION_REVISION);
		return 1;
	}
	if (retrievefromcommandline(argc,argv,"--bsd",NULL,0))
	{
		printf("Copyright 2019, dettus@dettus.net\n");
		printf("\n");
		printf("Redistribution and use in source and binary forms, with or without modification,\n");
		printf("are permitted provided that the following conditions are met:\n");
		printf("\n");
		printf("1. Redistributions of source code must retain the above copyright notice, this \n");
		printf("   list of conditions and the following disclaimer.\n");
		printf("\n");
		printf("2. Redistributions in binary form must reproduce the above copyright notice, \n");
		printf("   this list of conditions and the following disclaimer in the documentation \n");
		printf("   and/or other materials provided with the distribution.\n");
		printf("\n");
		printf("THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS \"AS IS\" AND\n");
		printf("ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED \n");
		printf("WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE \n");
		printf("DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE \n");
		printf("FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL \n");
		printf("DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR \n");
		printf("SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER \n");
		printf("CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, \n");
		printf("OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE \n");
		printf("OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n");
		printf("\n");
		return 1;


	}
	if (retrievefromcommandline(argc,argv,"--help",NULL,0))
	{
		printf("Plese run with %s -ini dMagnetic.ini GAME\n",argv[0]);
		printf(" -ini configuration file\n");
		printf("GAMES\n");
		printf("pawn\n");
		printf("guild\n");
		printf("jinxter\n");
		printf("corruption\n");
		printf("fish\n");
		printf("myth\n");
		printf("wonderland\n");
		printf(" OTHER PARAMETERS\n");
		printf(" --bsd shows the license\n");
		printf(" --help shows this help\n");
		printf(" --version shows %d.%d%d\n",VERSION_MAJOR,VERSION_MINOR,VERSION_REVISION);
		return 1;
	}
	if (retrievefromcommandline(argc,argv,"-ini",inifilename,sizeof(inifilename))) 
	{
		f_inifile=fopen(inifilename,"rb");
		if (!f_inifile) 
		{
			fprintf(stderr,"error opening %s\n",inifilename);
			return 0;
		}
	}
	if (retrievefromcommandline(argc,argv,"pawn",NULL,0))
	{
		magfilename[0]=gfxfilename[0]=0;
		retval=retrievefromini(f_inifile,"[FILES]","pawnmag",magfilename,sizeof(magfilename));
		retval=retrievefromini(f_inifile,"[FILES]","pawngfx",gfxfilename,sizeof(gfxfilename));
	}

	if (retrievefromcommandline(argc,argv,"guild",NULL,0))
	{
		magfilename[0]=gfxfilename[0]=0;
		retval=retrievefromini(f_inifile,"[FILES]","guildmag",magfilename,sizeof(magfilename));
		retval=retrievefromini(f_inifile,"[FILES]","guildgfx",gfxfilename,sizeof(gfxfilename));
	}

	if (retrievefromcommandline(argc,argv,"jinxter",NULL,0))
	{
		magfilename[0]=gfxfilename[0]=0;
		retval=retrievefromini(f_inifile,"[FILES]","jinxtermag",magfilename,sizeof(magfilename));
		retval=retrievefromini(f_inifile,"[FILES]","jinxtergfx",gfxfilename,sizeof(gfxfilename));
	}

	if (retrievefromcommandline(argc,argv,"corruption",NULL,0))
	{
		magfilename[0]=gfxfilename[0]=0;
		retval=retrievefromini(f_inifile,"[FILES]","corruptionmag",magfilename,sizeof(magfilename));
		retval=retrievefromini(f_inifile,"[FILES]","corruptiongfx",gfxfilename,sizeof(gfxfilename));
	}

	if (retrievefromcommandline(argc,argv,"fish",NULL,0))
	{
		magfilename[0]=gfxfilename[0]=0;
		retval=retrievefromini(f_inifile,"[FILES]","fishmag",magfilename,sizeof(magfilename));
		retval=retrievefromini(f_inifile,"[FILES]","fishgfx",gfxfilename,sizeof(gfxfilename));
	}

	if (retrievefromcommandline(argc,argv,"myth",NULL,0))
	{
		magfilename[0]=gfxfilename[0]=0;
		retval=retrievefromini(f_inifile,"[FILES]","mythmag",magfilename,sizeof(magfilename));
		retval=retrievefromini(f_inifile,"[FILES]","mythgfx",gfxfilename,sizeof(gfxfilename));
	}


	if (retrievefromcommandline(argc,argv,"wonderland",NULL,0))
	{
		magfilename[0]=gfxfilename[0]=0;
		retval=retrievefromini(f_inifile,"[FILES]","wonderlandmag",magfilename,sizeof(magfilename));
		retval=retrievefromini(f_inifile,"[FILES]","wonderlandgfx",gfxfilename,sizeof(gfxfilename));
	}


	///////////////////////////////////////////// init
	sharedmemsize=98304;
	pSharedMem=malloc(sharedmemsize);
	if (pSharedMem==NULL)
	{
		fprintf(stderr,"ERROR: unable to allocate shared memory\n");
		return 0;
	}
	retval=dMagnetic_init(&hVM68k,&hLineA,pSharedMem,sharedmemsize,magfilename,gfxfilename);
	if (retval)
	{
		return 0;
	}
	retval=default_getsize(&sizeGUI);
	if (retval)
	{
		fprintf(stderr,"ERROR. default_getsize returned %d\n",retval);
		return 0;
	}
	hGUI=malloc(sizeGUI);
	if (hGUI==NULL)
	{
		fprintf(stderr,"ERROR: unable to locate memory for the GUI\n");
		return 0;
	}
	retval=default_open(hGUI,f_inifile,argc,argv);
	if (retval)
	{
		fprintf(stderr,"ERROR: opening the GUI failed\n");
		return 0;
	}




	if (f_inifile) fclose(f_inifile);
	
	// set the call back hooks for this GUI
	retval|=lineA_setCBoutputChar(hLineA,default_cbOutputChar,	hGUI);
	retval|=lineA_setCBoutputString(hLineA,default_cbOutputString,	hGUI);
	retval|=lineA_setCBinputString(hLineA,default_cbInputString,	hGUI);
	retval|=lineA_setCBDrawPicture(hLineA,default_cbDrawPicture,	hGUI);
	retval|=lineA_setCBLoadGame(hLineA,default_cbLoadGame,		hGUI);
	retval|=lineA_setCBSaveGame(hLineA,default_cbSaveGame,		hGUI);
	if (retval)
	{
		fprintf(stderr,"ERROR: setting the API hooks failed\n");
		return 0;
	}
	/////////////////////////////////////////////		

	// everything is good. have fun playing!
	do
	{
		unsigned short opcode;
		unknownopcode=0;
		retval=vm68k_getNextOpcode(hVM68k,&opcode);
		if (retval==VM68K_OK) retval=lineA_substitute_aliases(hLineA,&opcode);
		if (retval==LINEA_OK) retval=vm68k_singlestep(hVM68k,opcode);
		if (retval!=VM68K_OK) retval=lineA_singlestep(hLineA,hVM68k,opcode);
		if (retval!=LINEA_OK && retval!=LINEA_OK_QUIT) 
		{
			fprintf(stderr,"\x1b[0m\n\x1b[0;37;44m Sorry. Unknown opcode %04X\x1b[0m\n",opcode);
			unknownopcode=1;
		}
//		if (retval==LINEA_OK_QUIT) printf("\x1b[0m\n\x1b[0;37;44mGoodbye!\x1b[0m\n");
	} while (!unknownopcode && !retval);	
	free(pSharedMem);
	free(hGUI);
	free(hLineA);
	free(hVM68k);
	return 1;
}
